from django.utils import translation
from django.utils.functional import lazy

from siba import SIBA_SETTINGS
from siba.locale import read_locale
from siba.parser import Parsable


class Default(dict):
    def __missing__(self, key):
        missing_parameter_handler = SIBA_SETTINGS.get("missing_parameter_handler")
        return missing_parameter_handler(key)


def inject_parameters(value: str, variables: dict):
    if variables is None:
        variables = {}

    return value.format_map(Default(variables))


def translate(key: str, prefix: str = None, locale: str = None, parameters: dict = None, **kwargs):
    prefix = SIBA_SETTINGS.get("default_prefix") if prefix is None else prefix
    locale = locale if locale is not None else translation.get_language()
    if locale is None:
        locale = SIBA_SETTINGS.get("default_locale")

    locale_content = read_locale(prefix, locale)

    try:
        unchecked_value = locale_content[key]

        if isinstance(unchecked_value, str):
            str_value = unchecked_value

        elif isinstance(unchecked_value, Parsable):
            parsable: Parsable = unchecked_value
            str_value = parsable.parse(key, **kwargs)

        else:
            raise Exception("Unexpected situation! shouldn't get here!")

        return inject_parameters(str_value, parameters)

    except KeyError:
        if SIBA_SETTINGS.get("error_on_unknown_key"):
            raise KeyError(
                f"Could not find key '{key}' in prefix '{prefix}' for locale '{locale}'"
            )
        return key


translate_lazy = lazy(translate, str)



from abc import ABC, abstractmethod


class Parsable(ABC):

    @abstractmethod
    def parse(self, key: str, **kwargs) -> str:
        pass


class ParsingRule(ABC):

    @abstractmethod
    def matches(self, data: dict) -> bool:
        pass

    @abstractmethod
    def parsed_value(self, data: dict) -> Parsable:
        pass




from os import path
from typing import Type

from siba import SIBA_SETTINGS
from siba.loader import FileLoader, LoaderStrategy
from siba.parsing_rules.pluralization import PluralizationParsingRule


PARSING_RULES = [PluralizationParsingRule()]


def format_file_content(content: dict, base_key: str = None, formatted: dict = None) -> dict:
    """
    From plain dictionary of translations we created a dictionary with dotted keys
    If values are supposed to be "decided on" dynamically per each call to translation then we register the parser
    If value is a string we don't check for parsers
    """
    if formatted is None:
        formatted = dict()

    for key, value in content.items():
        if base_key is None:
            ev_key = key
        else:
            ev_key = base_key + SIBA_SETTINGS['key_split_delimiter'] + key
        if isinstance(value, dict):
            matched = False
            for parsing_rule in PARSING_RULES:
                if parsing_rule.matches(value):
                    formatted[ev_key] = parsing_rule.parsed_value(value)
                    matched = True
                    break

            if not matched:
                format_file_content(value, ev_key, formatted)
        elif isinstance(value, str):
            formatted[ev_key] = value
        else:
            raise ValueError(f"Translation data should either be str or dictionary. Got {type(value)}")

    return formatted


__LOCALE_CACHE = {}


def _get_cache_key(prefix: str, locale: str):
    return f"{prefix}-{locale}"


# This function is not be thread-safe. If preload_locales is not explicitly called at application startup
# then this can lead to trying to cache locales multiple times
def read_locale(prefix: str, locale: str, loader_class: Type[FileLoader] = LoaderStrategy.JSON) -> dict:

    # If cache is enable check for the value in cache
    cache_key = _get_cache_key(prefix, locale)
    if SIBA_SETTINGS.get("cache_locales") and cache_key in __LOCALE_CACHE:
        return __LOCALE_CACHE.get(cache_key)

    # Read the file to load values
    loader = loader_class()
    content: dict = loader.load(
        path.join(SIBA_SETTINGS.get("locales_path"), f"{prefix}.{locale}.{loader.for_type()}"),
        SIBA_SETTINGS.get("error_on_missing_locale_file", True)
    )

    formatted_content = format_file_content(content)
    if SIBA_SETTINGS.get("cache_locales"):
        __LOCALE_CACHE[cache_key] = formatted_content

    return formatted_content


def preload_locales(loader_class: Type[FileLoader] = LoaderStrategy.JSON):
    """
    Preloads all locales into memory
    """
    if not SIBA_SETTINGS.get("cache_locales"):
        return

    loader = loader_class()
    for prefix in SIBA_SETTINGS.get("prefixes", []):
        for locale in SIBA_SETTINGS.get("locales", []):
            content = loader.load(
                path.join(SIBA_SETTINGS.get("locales_path"), f"{prefix}.{locale}.{loader_class.for_type()}"),
                False
            )
            if content is None:
                continue

            formatted_content = format_file_content(content)
            __LOCALE_CACHE[_get_cache_key(prefix, locale)] = formatted_content


def clear_cached_locales():
    __LOCALE_CACHE.clear()


def get_cached_locales():
    return __LOCALE_CACHE



import logging
from typing import Any

logger = logging.getLogger(__name__)


SIBA_SETTINGS = {
    "error_on_unknown_key": False,
    "error_on_missing_locale_file": True,
    "key_split_delimiter": ".",
    "locales_path": None,
    "cache_locales": True,
    "prefixes": ["application"],
    "default_prefix": "application",
    "default_locale": "en",
    "locales": ["en"],
    "missing_parameter_handler": lambda x: x,
    "pluralization": {
        "some_enabled": True,
        "some_limit": 4
    }
}


def set_setting(key: str, value: Any):
    if key not in SIBA_SETTINGS:
        logger.warning(f"Siba settings don't seem to expect the key")

    SIBA_SETTINGS[key] = value


def get_setting(key: str):
    return SIBA_SETTINGS.get(key)

